//----------------------------------------------------------------------------
// hash tables for spell checking
// copyright 2010 christopher rasmussen
// university of delaware
//----------------------------------------------------------------------------

#include <iostream>
#include <ctype.h>
#include <fstream>
#include <vector>
#include <cstdlib>
#include <string>
#include <math.h>
#include "sys/time.h"

using namespace std;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

const float MAX_LOAD = 1.0;

class String_HT
{
public:

  String_HT(int);
  int hash_index(string);                           // hash function
  int next_prime_greater_than_or_equal_to(int);

  virtual void insert(string) = 0;    
  virtual void remove(string) = 0;
  virtual bool find(string) = 0;
  virtual void expand_and_rehash() = 0;
  virtual void print() = 0;
  virtual void gather(vector <string> &) = 0;
  virtual void remove_all() = 0;

  int table_size;
  int num_entries;
  float load;

};

//----------------------------------------------------------------------------

// collisions resolved using SEPARATE CHAINING

class Chain_String_HT : public String_HT
{
public:

  Chain_String_HT(int);

  void insert(string);    
  void remove(string);
  bool find(string);
  void expand_and_rehash();
  void print();
  void gather(vector <string> &);
  void remove_all();

  vector < vector <string> > table;      // the hash table itself
  
};

//----------------------------------------------------------------------------

// collisions resolved using QUADRATIC PROBING

class Probe_String_HT : public String_HT
{
public:

  Probe_String_HT(int);

  void insert(string);    
  void remove(string);
  bool find(string);
  void expand_and_rehash();
  void print();
  void gather(vector <string> &);
  void remove_all();
   
  // ???                          // the hash table itself
  
};

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// initialize hash table...
// sets table size to smallest prime >= num_elements

String_HT::String_HT(int num_elements)
{
  table_size = next_prime_greater_than_or_equal_to(num_elements);

  cout << num_elements << " -> " << table_size << endl;

  num_entries = 0;
  load = 0.0;
}

//----------------------------------------------------------------------------

// given an integer n, what is the smallest prime number p
// such that p >= n ?

// does this the dumbest possible way :)

int String_HT::next_prime_greater_than_or_equal_to(int n)
{
  int i, sqrt_n, p;

  // search over interval where we know there must be a prime

  for (p = n; p < 2 * n; p++) {

    sqrt_n = (int) floor(sqrt(n));

    // check all possible divisors of n

    for (i = 2; i <= sqrt_n; i++) 
      if ((p % i) == 0)   // found a divisor...keep looking
	break;

    // made it to end with no divisor => p is prime!

    if (i == sqrt_n + 1) 
      return p;
  }
}

//----------------------------------------------------------------------------

// THE HASH FUNCTION...takes key to table index using
// polynomial scheme, computed with Horner's method 
// (see pp. 187-188 of Weiss)

int String_HT::hash_index(string key)
{
  int value = 0;

  for (int i = 0; i < key.size(); i++)
    value = 37 * value + key[i];

  value %= table_size;

  if (value < 0) 
    value += table_size;

  return value;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// constructor...assumes parent class constructor sets table_size from num_elements

Chain_String_HT::Chain_String_HT(int num_elements) : String_HT(num_elements)
{
  table.resize(table_size);

  remove_all();
}

//----------------------------------------------------------------------------

// print entire hash table--i.e., each entry of each chain, including empty ones

void Chain_String_HT::print()
{
  printf("load = %.3f\n", load);

  for (int i = 0; i < table.size(); i++) {
    cout << i << " (" << table[i].size() << "): ";
    for (int j = 0; j < table[i].size(); j++) 
      cout << table[i][j] << " ";
    cout << endl;
  }
}

//----------------------------------------------------------------------------

// put key into table.  if load is too big, expand and rehash

void Chain_String_HT::insert(string key)
{
  // FILL IN FROM HERE...

  // ???

  // ...TO HERE

  // update load stats and expand table if necessary

  num_entries++;
  load = (float) num_entries / (float) table.size();

  if (load >= MAX_LOAD) 
    expand_and_rehash();

}

//----------------------------------------------------------------------------

// double size of hash table to reduce load

void Chain_String_HT::expand_and_rehash()
{
  int i, j;
  vector <string> V;

  // move all stored keys to a temporary vector 

  gather(V);
  remove_all();

  // hash table is now in original (empty) state; 
  // reallocate hash table at twice the size

  table_size = next_prime_greater_than_or_equal_to(2 * table.size());

  table.resize(table_size);

  num_entries = 0;

  // reinsert all keys

  while (!V.empty()) {
    insert(V.back());
    V.pop_back();
  }

  // recalculate load

  load = (float) num_entries / (float) table.size();
}

//----------------------------------------------------------------------------

// FILL THIS IN...

// take key out of hash table if it's there

void Chain_String_HT::remove(string key)
{
  // ???
}

//----------------------------------------------------------------------------

// FILL IN

// is key in hash table?

bool Chain_String_HT::find(string key)
{
  // ???
}

//----------------------------------------------------------------------------

// collect all non-empty entries in hash table into vector...
// does not affect hash table structure

void Chain_String_HT::gather(vector <string> & V)
{
  V.clear();

  for (int i = 0; i < table.size(); i++) 
    for (int j = 0; j < table[i].size(); j++)
      V.push_back(table[i][j]);
}

//----------------------------------------------------------------------------

// make hash table empty without affecting table size

void Chain_String_HT::remove_all()
{
  for (int i = 0; i < table.size(); i++) 
    table[i].clear();
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// HONORS section:

// FILL IN function definitions of QUADRATIC PROBING versions of the following here:

//   void insert(string);    
//   void remove(string);
//   bool find(string);
//   void expand_and_rehash();
//   void print();
//   void gather(vector <string> &);
//   void remove_all();

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

class SpellChecker
{
public:

  SpellChecker(string);
  void find_bad_words(string);
  void suggest_good_words();
  void cleanup_string(string &);
  void split_string(string &, string &);
  void print_suggestions(string);

  Chain_String_HT *Dictionary;  
  Chain_String_HT *BadWords;  
  
};

//----------------------------------------------------------------------------

SpellChecker::SpellChecker(string dictionary_name)
{
  // initialize underlying hash table data structures

  Dictionary = new Chain_String_HT(1000);
  BadWords = new Chain_String_HT(1000);

  // open file and read

  ifstream inStream;
  string s;

  inStream.open(dictionary_name.c_str());

  if (inStream.fail()) {
    cout << "Failed to open file\n";
    exit(1);
  }

  while (!inStream.eof()) {

    inStream >> s;

    // write to file and echo to console
    
    if (!inStream.eof()) {

      cleanup_string(s); 

      if (s.size())
	Dictionary->insert(s);

      // echo what we just put in 
      //      cout << s << endl;
    }

  }

  inStream.close();

}

//----------------------------------------------------------------------------

// make string all lower-case, remove leading/trailing punctuation, etc.

void SpellChecker::cleanup_string(string & s)
{
  int i;

  while (s.size() > 0 && ispunct(s[0]))
    s.erase(0, 1);

  while (s.size() > 0 && ispunct(s[s.size() - 1]))
    s.erase(s.size() - 1, 1);

  for (i = 0; i < s.size(); i++)
    if (isalpha(s[i]))
      s[i] = tolower(s[i]);
}

//----------------------------------------------------------------------------

// assuming s is already "clean"...now just see if there's punctuation in the
// MIDDLE of the word that (besides one apostrophe) that may be dividing two words

void SpellChecker::split_string(string & s, string & t)
{
  int i, j;

  t.clear();

  // get to non-alpha part of string s

  for (i = 1; i < s.size() && (isalpha(s[i]) || s[i] == '\''); i++)
    ;

  // erase it

  if (i < s.size() - 1) {
    while (i < s.size() - 1 && !isalpha(s[i]))
      s.erase(i, 1);
  }

  // put remaining alpha part of string into t

  while (i < s.size()) {
    t.push_back(s[i]);
    s.erase(i, 1);
  }
}

//----------------------------------------------------------------------------

// FILL IN

// list, indented, any string mutations that make s a dictionary word or 
// else say "NONE"

void SpellChecker::print_suggestions(string s)
{
  // ???
}

//----------------------------------------------------------------------------

void SpellChecker::find_bad_words(string filename)
{
  // open filename

  ifstream inStream;

  inStream.open(filename.c_str());

  if (inStream.fail()) {
    cout << "Failed to open file\n";
    exit(1);
  }

  // iterate through file and check each word

  string s, t;

  while (!inStream.eof()) {

    inStream >> s;

    if (!inStream.eof()) {

      cleanup_string(s); 

      if (s.size()) {

	split_string(s, t);   // see if s has punctuation in middle

	if (!Dictionary->find(s)) 
	  BadWords->insert(s);

	// this is only true if split_string() did something

	if (t.size()) {

	  if (!Dictionary->find(t)) 
	    BadWords->insert(t);

	}
      }
    }

  }

  inStream.close();

}

//----------------------------------------------------------------------------

void SpellChecker::suggest_good_words()
{
  vector <string> V;

  BadWords->gather(V);

  for (int i = 0; i < V.size(); i++) {
    cout << i << ": " << V[i] << " ... suggestions:\n";
    print_suggestions(V[i]);
  }
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

int main(int argc, char **argv)
{
  if (argc < 2) {
    cout << "hashcheck <filename>\n";
    exit(1);
  }

  SpellChecker SC("words.txt");

  SC.find_bad_words(argv[1]);

  SC.suggest_good_words();

  return 1;
}
